/**
 * textstream.c
 *
 * @author Martin Hrlin
 */

/**	
 * This software was developed with funding from the National Institute 
 * on Disability and Rehabilitation Research, U.S. Department of Education, 
 * grant number H133E090001 (RERC on Telecommunications Access). However, 
 * those contents do not necessarily represent the policy of the Department 
 * of Education, and you should not assume endorsement by the Federal Government. 
 */

#include "linphonecore.h"
#include "textstream.h"

//#define TS_DEBUG

#define TS_OUTBUF_SIZE 32
#define TS_REDGEN 2
#define TS_NUMBER_OF_OUTBUF TS_REDGEN+1
#define TS_INBUF_SIZE TS_OUTBUF_SIZE*TS_NUMBER_OF_OUTBUF
#define TS_KEEP_ALIVE_INTERVAL 10000
#define TS_SEND_INTERVAL 299

#define TS_FLAG_NOTFIRST 0x01
#define TS_FLAG_NOCALLBACK 0x02

#ifdef TS_DEBUG
#define ts_debug(_fmt,...) ms_message("textstream.c[%i]:" _fmt, __LINE__,##__VA_ARGS__)
static void ts_hexdump(const uint8_t* data, const size_t s){
	uint8_t buf[1000];
	const uint8_t* c;
	c = data;
	buf[0] = '\0';
	int pos = 0;
	int i = 0;
	while(c<data+s){
		pos+=snprintf(&buf[pos],1000-pos," ['%c']0x%X",*c>20?*c:'X',*c);
		c++;
		i++;
	}
	buf[pos]='\0';
	ts_debug("%s", buf);
}
#else
#define ts_debug(_fmt,...) (void)0
#define ts_hexdump(a,b) (void)0
#endif
#define ts_warning(_fmt,...) ms_warning("textstream.c[%i]:" _fmt, __LINE__,##__VA_ARGS__)

struct _TextStream{
	int flags;
	RtpSession *session;
	int pt_t140;
	int pt_red;	
	int prevseqno;
	uint8_t inbuf[TS_INBUF_SIZE];
	size_t inbufsize;
	uint8_t* inbufpos;
	uint8_t buf[TS_NUMBER_OF_OUTBUF][TS_OUTBUF_SIZE];
	int pribuf;
	size_t bufsize[TS_NUMBER_OF_OUTBUF];
	uint32_t timestamp[TS_NUMBER_OF_OUTBUF];
};

extern RtpSession *create_duplex_rtpsession(int locport,bool_t ipv6);

static uint32_t gettime() {
	struct timeval t;
	gettimeofday(&t,NULL);
	return (t.tv_sec*1000+t.tv_usec/1000);
}

/**
 * How many bytes makes up one character.
 **/
static int utf8_test(const uint8_t c) {
	if (!(c&0x80)){ return 1; }
	else if (!(c&0x40)){ return 0; }
	else if (!(c&0x20)){ return 2; }
	else if (!(c&0x10)){ return 3; }
	else if (!(c&0x08)){ return 4; }
	else { return -1; }
}

static int validate_utf8_buf(const uint8_t* b, const size_t s) {
	int i,t;
	for (i=0,t=0;i<s;i++,t--) {
		if (t==0) {
			t=utf8_test(b[i]);
			if (t<=0) {
				return 0;
			}
		} else {
			if (utf8_test(b[i])!=0) {
				return 0;
			}
		}
	}
	if (t) {
		return 0;
	}
	return 1; /* SUCCESS */
}

static int red_needed(int cur, int prev){
	int t = cur-prev;
	if (t>0) {
		return t-1;
	} else if (t<-100) {
		return t+0xFFFF;
	}
	return -1;
}

void text_stream_putchar(TextStream *ts,const char c){
	int i=ts->pribuf;
	if (ts->bufsize[i]<TS_OUTBUF_SIZE) {
		ts->buf[i][ts->bufsize[i]++]=c;
	} else {
		ts_warning("Text stream out buf is full, can not add more characters to buffer!");
	}
}

static inline int is_data_to_send(const TextStream *ts){
	int i;
	for (i=0;i<=TS_REDGEN;i++){
		if (ts->bufsize[i]){
			return 1;
		}
	}
	return 0;
}

static inline int get_last_buf(const TextStream *ts){
	return ts->pribuf==TS_REDGEN ? 0 : ts->pribuf+1;
}

static inline int get_next_buf(const TextStream *ts, const int cur){
	if (cur==ts->pribuf){
		return -1;
	}
	return cur==TS_REDGEN ? 0 : cur+1;
}

static void use_next_buf(TextStream *ts){
	ts->pribuf=get_last_buf(ts);
	ts->bufsize[ts->pribuf]=0;
	ts->timestamp[ts->pribuf]=0;
}

static inline uint32_t get_prev_time(const TextStream *ts){
	return ts->timestamp[ts->pribuf ? ts->pribuf-1 : TS_REDGEN];
}

static inline uint32_t get_red_subheader(const int pt, const int offset, const 
										 int length){
	return (1<<31)|((0x7F&pt)<<24)|((0x3FFF&offset)<<10)|(0x3FF&length);
}

static mblk_t *text_stream_generate_red_packet(TextStream *ts){
	int cur=get_last_buf(ts);
	uint8_t t140=(ts->pt_t140&0x7F);
	int pri=ts->pribuf;
	mblk_t *packet;
	uint8_t payload[TS_OUTBUF_SIZE*TS_NUMBER_OF_OUTBUF+4*TS_REDGEN+1];
	int32_t payloadsize=0;
	int mark=1;

	/* this makes it possible to generate t140 with same function... */
	if(ts->pt_red>0){
		while (cur!=pri){
			uint32_t diff=ts->timestamp[pri]-ts->timestamp[cur];
			int size=ts->bufsize[cur];
			uint32_t sub=htonl(get_red_subheader(t140,diff,size));
			
			memcpy(&payload[payloadsize],&sub,4);
			payloadsize+=4;
			
			cur=get_next_buf(ts,cur);
		}
		memcpy(&payload[payloadsize],&t140,1);
		payloadsize+=1;
		
		cur=get_last_buf(ts);
		while(cur!=pri){
			if(ts->bufsize[cur]){
				mark=0;
				memcpy(&payload[payloadsize],&ts->buf[cur][0],ts->bufsize[cur]);
				payloadsize+=ts->bufsize[cur];
			}
			cur=get_next_buf(ts,cur);
		}
	}
	if(ts->bufsize[pri]){
		memcpy(&payload[payloadsize],&ts->buf[pri][0],ts->bufsize[pri]);
		payloadsize+=ts->bufsize[pri];
	}
	packet=rtp_session_create_packet(ts->session,RTP_FIXED_HEADER_SIZE,payload,
									 payloadsize);
	rtp_set_markbit(packet,mark);
	return packet;
}

static void send_data(TextStream *ts){
	uint32_t time=gettime();
	int i = ts->pribuf;
	uint32_t prevtime = get_prev_time(ts);
	if (is_data_to_send(ts)) {
		if (time<prevtime||time-prevtime>TS_SEND_INTERVAL) {
			ts->timestamp[i] = time;
			rtp_session_sendm_with_ts(ts->session,
									  text_stream_generate_red_packet(ts),time);
			use_next_buf(ts);
		}
	}else{
		if (time<prevtime||time-prevtime>TS_KEEP_ALIVE_INTERVAL){
			text_stream_putchar32(ts,0xFEFF); /* BOM */
			send_data(ts);
		}
	}
}

TextStream *text_stream_new(int locport,bool_t ipv6){
	TextStream *ts=(TextStream *)ms_new0(TextStream,1);
	ts->session=create_duplex_rtpsession(locport,ipv6);
	rtp_session_enable_adaptive_jitter_compensation(ts->session,0);
	rtp_session_enable_jitter_buffer(ts->session,0);
	ts->inbufpos=ts->inbuf;
	return ts;
}

void text_stream_start(TextStream *ts,RtpProfile *profile,const char *rem_addr,
					   int rem_port,int rem_rtcp_port,int pt /*ignored*/){
	ts_debug("before crash?");
	rtp_session_set_profile(ts->session,profile);
	if (rem_port>0) {
		rtp_session_set_remote_addr_full(ts->session,rem_addr,rem_port,
										 rem_rtcp_port);
	}

	ts->pt_t140=rtp_profile_get_payload_number_from_mime(profile,"t140");
	ts->pt_red=rtp_profile_get_payload_number_from_mime(profile,"red");
	
	if (ts->pt_red>0&&ts->pt_t140>0){
		rtp_session_set_payload_type(ts->session,ts->pt_red);
	}else if (ts->pt_t140>0){
		rtp_session_set_payload_type(ts->session,ts->pt_t140);
	}else{
		/* we dont know this kind of textstream... */
		ts_warning("unkown type of textstream");
	}
}

void text_stream_stop(TextStream *ts){
	if(ts->session!=NULL){
		rtp_session_destroy(ts->session);
	}
	ms_free(ts);
}

static void insert_lost_char(uint8_t *p){
	p[0] = 0xEF;
	p[1] = 0xBF;
	p[2] = 0xBD;
}

static int read_t140_data(TextStream *ts, uint8_t *data, int readsize){
	if (readsize<0) {
		ts_warning("corrupt packet (readsize<0)");
		return -1;
	}else if (readsize>TS_INBUF_SIZE-ts->inbufsize) {
		readsize=TS_INBUF_SIZE-ts->inbufsize;
		ts_warning("reading less characters than in buffer");
		
		/*POTENTIAL BUG... if last char is a multi char but just parts of it get 
		in, next function fails and whole buf is ignored. */
	}
	if (readsize>0) {
		ts_debug("Reading %i bytes\n",readsize);
		if (!validate_utf8_buf(data,readsize)){
			ts_hexdump(data,readsize);
			ts_warning("not a valid utf8 payload");
			ts->inbufsize=0;
			return -1;
		}
		memcpy(&ts->inbuf[ts->inbufsize],data,readsize);
		ts->inbufsize+=readsize;
	}
	return 0; /* 0 on success */
}

static void process_t140_packet(TextStream *ts,mblk_t *packet){
	int seqno=rtp_get_seqnumber(packet);
	uint8_t *payload;
	int payloadsize=rtp_get_payload(packet,&payload);
	
	ts_debug("t140 seqno:%i",seqno);
	if(ts->flags&TS_FLAG_NOTFIRST){
		int t = red_needed(seqno,ts->prevseqno);
		if (t<0) {
			ts_warning("packet arrived out of order");
			return;
		} else if (t>0) {
			ts->inbufsize=3;
			insert_lost_char(ts->inbuf);
		}
	}
	if (read_t140_data(ts,payload,payloadsize)) {
		return; /* return without updatting seqno */
	}
	ts->prevseqno=seqno;
}

static void process_red_packet(TextStream *ts,mblk_t *packet){
	int seqno=rtp_get_seqnumber(packet);
	uint8_t *payload;
	int redgen=0;
	int pos=0;
	int redneeded;
	int readstart;
	int payloadsize=rtp_get_payload(packet,&payload);
	/* check how many is red, also check if its the right payload for the red */
	
	ts_debug("red seqno:%i",seqno);
	while((pos<payloadsize)&&(payload[pos]&(1<<7))){
		redgen++;
		if(((int)payload[pos]&0x7F)!=ts->pt_t140){
			ts_warning("invalid red packet");
			return;
		}
		pos+=4;
	}
	
	ts_debug("red redgen:%i",redgen);
	if((int)payload[pos]!=ts->pt_t140){
		ts_warning("invalid red packet");
		return;
	}
	if (ts->flags&TS_FLAG_NOTFIRST) {
		redneeded=red_needed(seqno,ts->prevseqno);
	} else {
		redneeded=0;
	}
	if (redneeded<0) {
		ts_warning("packet arrived out of order");
		return;
	}
	
	ts_debug("red redneeded:%i",redneeded);
	if (redneeded>redgen){
		/* we need more red than we got */
		ts->inbufsize=3;
		insert_lost_char(ts->inbuf);
		redneeded=redgen;
	}
	/* loop over unneeded red */
	readstart=redgen*4+1;
	for (pos=0;pos<(redgen-redneeded)*4;pos+=4){
		readstart+=(((uint32_t)payload[pos+2]<<8)|(uint32_t)payload[pos+3])&0x3FF;
	}
	if (read_t140_data(ts,&payload[readstart],payloadsize-readstart)) {
		ts_debug("error reading");
		return; /* return without updating seqno */
	}
	ts->prevseqno=seqno;
}

static int read_text_packet(TextStream *ts){
	uint32_t time=gettime();
	mblk_t *packet=rtp_session_recvm_with_ts(ts->session,time);
	int pt;
	
	ts->inbufpos=ts->inbuf;
	ts->inbufsize=0;
	
	if (packet==NULL){
		return 0;
	}
	pt=rtp_get_payload_type(packet);
	if (pt==ts->pt_t140){
		process_t140_packet(ts,packet);
	}else if (ts->pt_red&&pt==ts->pt_red){
		process_red_packet(ts,packet);
	}else {
		ts_warning("unkown pt for text packet");
		ts_debug("p:%i t:%i r:%i",pt,ts->pt_t140,ts->pt_red);
	}
	if (!(ts->flags&TS_FLAG_NOTFIRST)) {
		ts->flags|=TS_FLAG_NOTFIRST;
	}
	ortp_free(packet);
	return 1;
}

char text_stream_getchar(TextStream *ts){
	uint8_t *p = ts->inbufpos;
	uint8_t *end = &ts->inbuf[ts->inbufsize];
	while (end>p) {
		if(p[0] != '\0') {
			if (end-p>=3) {
				if (p[0]==0xEF && p[1]==0xBB && p[2]==0xBF){ /* BOM */
					p+=3;
					continue;
				}
			}
			ts->inbufpos = p+1;
			return *p;
		}
		p++;
	}
	if(read_text_packet(ts)){
		return text_stream_getchar(ts);
	}
	ts->flags&=~TS_FLAG_NOCALLBACK;
	return '\0';
}

uint32_t text_stream_getchar32(TextStream *ts) {
	uint32_t c = text_stream_getchar(ts);
	int t = utf8_test(c);
	switch (t) {
		case 1:
			return c;
		case 2:
			c = (c&0x1F)<<6;
			c+= ((uint32_t)text_stream_getchar(ts)&0x3F);
			return c;
		case 3:
			c = (c&0x0F)<<12;
			c+= (((uint32_t)text_stream_getchar(ts)&0x3F)<<6);
			c+= ((uint32_t)text_stream_getchar(ts)&0x3F);
			return c;
		case 4:
			c = (c&0x7)<<19;
			c+= (((uint32_t)text_stream_getchar(ts)&0x3F)<<12);
			c+= (((uint32_t)text_stream_getchar(ts)&0x3F)<<6);
			c+= ((uint32_t)text_stream_getchar(ts)&0x3F);
			return c;
		default:
			return 0;
	}
}

static int text_stream_ischar(TextStream *ts){
	if (ts->inbufsize) {
		return 1;
	}
	return read_text_packet(ts);
}

int text_stream_iterate(TextStream *ts){
	send_data(ts);
	if (ts->flags&TS_FLAG_NOCALLBACK){
		return 0;
	}
	if (text_stream_ischar(ts)){
		ts->flags|=TS_FLAG_NOCALLBACK;
		return 1;
	}
	return 0;
}

void text_stream_putchar32(TextStream *ts,uint32_t ic){
	int i=ts->pribuf;
	uint8_t* c = &ts->buf[i][ts->bufsize[i]];
	if (ic<0x80) {
		if (ts->bufsize[i]<TS_OUTBUF_SIZE){
			c[0] = ic;
			ts->bufsize[i]++;
		}
	} else if (ic<0x800) {
		if (ts->bufsize[i]+1<TS_OUTBUF_SIZE){
			c[1] = 0x80+((ic&0x3F));
			c[0] = 0xC0+((ic>>6)&0x1F);
			ts->bufsize[i]+=2;
		}
	} else if (ic<0x100000) {
		if (ts->bufsize[i]+2<TS_OUTBUF_SIZE){
			c[2] = 0x80+(ic&0x3F);
			c[1] = 0x80+((ic>>6)&0x3F);
			c[0] = 0xE0+((ic>>12)&0xF);
			ts->bufsize[i]+=3;
		}
	} else if (ic<0x110000) {
		if (ts->bufsize[i]+3<TS_OUTBUF_SIZE){
			c[3] = 0x80+(ic&0x3F);
			c[2] = 0x80+((ic>>6)&0x3F);
			c[1] = 0x80+((ic>>12)&0x3F);
			c[0] = 0xF0+((ic>>18)&0x7);
			ts->bufsize[i]+=4;
		}
	}
}
